# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from buildbot.util import runprocess


def _check_env_is_expected(test, expected_env, env):
    if expected_env is None:
        return

    env = env or {}
    for var, value in expected_env.items():
        test.assertEqual(env.get(var), value,
                         'Expected environment to have {} = {}'.format(var, repr(value)))


class ExpectMaster:
    _stdout = b""
    _stderr = b""
    _exit = 0
    _workdir = None
    _env = None

    def __init__(self, command):
        self._command = command

    def stdout(self, stdout):
        assert(isinstance(stdout, bytes))
        self._stdout = stdout
        return self

    def stderr(self, stderr):
        assert(isinstance(stderr, bytes))
        self._stderr = stderr
        return self

    def exit(self, exit):
        self._exit = exit
        return self

    def workdir(self, workdir):
        self._workdir = workdir
        return self

    def env(self, env):
        self._env = env
        return self

    def check(self, test, command, workdir, env):
        test.assertDictEqual({'command': command, 'workdir': workdir},
                             {'command': self._command, 'workdir': self._workdir},
                             "unexpected command run")

        _check_env_is_expected(test, self._env, env)
        return (self._exit, self._stdout, self._stderr)

    def __repr__(self):
        return "<ExpectMaster(command={})>".format(self._command)


class MasterRunProcessMixin:
    long_message = True

    def setup_master_run_process(self):
        self._master_run_process_patched = False
        self._expected_master_commands = []
        self._master_run_process_expect_env = {}

    def assert_all_commands_ran(self):
        self.assertEqual(self._expected_master_commands, [],
                         "assert all expected commands were run")

    def patched_run_process(self, reactor, command, workdir=None, env=None,
                            collect_stdout=True, collect_stderr=True, stderr_is_error=False,
                            io_timeout=300, runtime_timeout=3600, sigterm_timeout=5,
                            initial_stdin=None):

        _check_env_is_expected(self, self._master_run_process_expect_env, env)

        if not self._expected_master_commands:
            self.fail("got command {} when no further commands were expected".format(command))

        expect = self._expected_master_commands.pop(0)

        rc, stdout, stderr = expect.check(self, command, workdir, env)

        if not collect_stderr and stderr_is_error and stderr:
            rc = -1

        if collect_stdout and collect_stderr:
            return (rc, stdout, stderr)
        if collect_stdout:
            return (rc, stdout)
        if collect_stderr:
            return (rc, stderr)
        return rc

    def _patch_runprocess(self):
        if not self._master_run_process_patched:
            self.patch(runprocess, "run_process", self.patched_run_process)
            self._master_run_process_patched = True

    def add_run_process_expect_env(self, d):
        self._master_run_process_expect_env.update(d)

    def expect_commands(self, *exp):
        for e in exp:
            if not isinstance(e, ExpectMaster):
                raise Exception('All expectation must be an instance of ExpectMaster')

        self._patch_runprocess()
        self._expected_master_commands.extend(exp)



# This file is part of Buildbot. Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import sqlalchemy as sa

from buildbot.util import sautils


def upgrade(migrate_engine):
    metadata = sa.MetaData()
    metadata.bind = migrate_engine

    steps = sautils.Table('steps', metadata, autoload=True)

    idx = sa.Index('steps_started_at',
                   steps.c.started_at)
    idx.create()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re

from buildbot.util import ComparableMixin
from buildbot.util import NotABranch


def is_re_pattern(obj):
    # re.Pattern only exists in Python 3.7
    return hasattr(obj, 'search') and hasattr(obj, 'match')


def extract_filter_values(values, filter_name):
    if not isinstance(values, (list, str)):
        raise ValueError("Values of filter {} must be list of strings or a string".format(
            filter_name))
    if isinstance(values, str):
        values = [values]
    else:
        for value in values:
            if not isinstance(value, str):
                raise ValueError("Value of filter {} must be string".format(filter_name))
    return values


def extract_filter_values_branch(values, filter_name):
    if not isinstance(values, (list, str, type(None))):
        raise ValueError("Values of filter {} must be list of strings, a string or None".format(
            filter_name))
    if isinstance(values, (str, type(None))):
        values = [values]
    else:
        for value in values:
            if not isinstance(value, (str, type(None))):
                raise ValueError("Value of filter {} must be string or None".format(filter_name))
    return values


def extract_filter_values_regex(values, filter_name):
    if not isinstance(values, (list, str)) and not is_re_pattern(values):
        raise ValueError("Values of filter {} must be list of strings, a string or regex".format(
            filter_name))
    if isinstance(values, str) or is_re_pattern(values):
        values = [values]
    else:
        for value in values:
            if not isinstance(value, str) and not is_re_pattern(value):
                raise ValueError("Value of filter {} must be string or regex".format(filter_name))
    return values


class _FilterExactMatch:
    def __init__(self, values):
        self.values = values

    def is_matched(self, value):
        return value in self.values

    def describe(self, prop):
        return '{} in {}'.format(prop, self.values)


class _FilterExactMatchInverse:
    def __init__(self, values):
        self.values = values

    def is_matched(self, value):
        return value not in self.values

    def describe(self, prop):
        return '{} not in {}'.format(prop, self.values)


class _FilterRegex:
    def __init__(self, regexes):
        self.regexes = [self._compile(regex) for regex in regexes]

    def _compile(self, regex):
        if is_re_pattern(regex):
            return regex
        return re.compile(regex)

    def is_matched(self, value):
        if value is None:
            return False
        for regex in self.regexes:
            if regex.match(value) is not None:
                return True
        return False

    def describe(self, prop):
        return '{} matches {}'.format(prop, self.regexes)


class _FilterRegexInverse:
    def __init__(self, regexes):
        self.regexes = [self._compile(regex) for regex in regexes]

    def _compile(self, regex):
        if is_re_pattern(regex):
            return regex
        return re.compile(regex)

    def is_matched(self, value):
        if value is None:
            return True
        for regex in self.regexes:
            if regex.match(value) is not None:
                return False
        return True

    def describe(self, prop):
        return '{} does not match {}'.format(prop, self.regexes)


class SourceStampFilter(ComparableMixin):

    compare_attrs = (
        'filter_fn',
        'project_filters',
        'codebase_filters',
        'repository_filters',
        'branch_filters'
    )

    def __init__(self,
                 # gets a SourceStamp dictionary, returns boolean
                 filter_fn=None,

                 project_eq=None, project_not_eq=None, project_re=None, project_not_re=None,
                 repository_eq=None, repository_not_eq=None,
                 repository_re=None, repository_not_re=None,
                 branch_eq=NotABranch, branch_not_eq=NotABranch, branch_re=None, branch_not_re=None,
                 codebase_eq=None, codebase_not_eq=None, codebase_re=None, codebase_not_re=None):

        self.filter_fn = filter_fn
        self.project_filters = self.create_filters(project_eq, project_not_eq,
                                                   project_re, project_not_re, 'project')
        self.codebase_filters = self.create_filters(codebase_eq, codebase_not_eq,
                                                    codebase_re, codebase_not_re, 'codebase')
        self.repository_filters = self.create_filters(repository_eq, repository_not_eq,
                                                      repository_re, repository_not_re,
                                                      'repository')
        self.branch_filters = self.create_branch_filters(branch_eq, branch_not_eq,
                                                         branch_re, branch_not_re, 'branch')

    def create_branch_filters(self, eq, not_eq, regex, not_regex, filter_name):
        filters = []
        if eq is not NotABranch:
            values = extract_filter_values_branch(eq, filter_name + '_eq')
            filters.append(_FilterExactMatch(values))

        if not_eq is not NotABranch:
            values = extract_filter_values_branch(not_eq, filter_name + '_not_eq')
            filters.append(_FilterExactMatchInverse(values))

        if regex is not None:
            values = extract_filter_values_regex(regex, filter_name + '_re')
            filters.append(_FilterRegex(values))

        if not_regex is not None:
            values = extract_filter_values_regex(not_regex, filter_name + '_re')
            filters.append(_FilterRegexInverse(values))

        return filters

    def create_filters(self, eq, not_eq, regex, not_regex, filter_name):
        filters = []
        if eq is not None:
            values = extract_filter_values(eq, filter_name + '_eq')
            filters.append(_FilterExactMatch(values))

        if not_eq is not None:
            values = extract_filter_values(not_eq, filter_name + '_not_eq')
            filters.append(_FilterExactMatchInverse(values))

        if regex is not None:
            values = extract_filter_values_regex(regex, filter_name + '_re')
            filters.append(_FilterRegex(values))

        if not_regex is not None:
            values = extract_filter_values_regex(not_regex, filter_name + '_re')
            filters.append(_FilterRegexInverse(values))

        return filters

    def do_prop_match(self, ss, prop, filters):
        value = ss.get(prop, '')
        for filter in filters:
            if not filter.is_matched(value):
                return False
        return True

    def is_matched(self, ss):
        if self.filter_fn is not None and not self.filter_fn(ss):
            return False
        if self.project_filters and not self.do_prop_match(ss, 'project', self.project_filters):
            return False
        if self.codebase_filters and not self.do_prop_match(ss, 'codebase', self.codebase_filters):
            return False
        if self.repository_filters and \
                not self.do_prop_match(ss, 'repository', self.repository_filters):
            return False
        if self.branch_filters and not self.do_prop_match(ss, 'branch', self.branch_filters):
            return False
        return True

    def is_matched_codebase(self, codebase):
        for filter in self.codebase_filters:
            if not filter.is_matched(codebase):
                return False
        return True

    def _repr_filters(self, filters, prop):
        return [filter.describe(prop) for filter in filters]

    def __repr__(self):
        filters = []
        if self.filter_fn is not None:
            filters.append('{}()'.format(self.filter_fn.__name__))
        filters += self._repr_filters(self.project_filters, 'project')
        filters += self._repr_filters(self.codebase_filters, 'codebase')
        filters += self._repr_filters(self.repository_filters, 'repository')
        filters += self._repr_filters(self.branch_filters, 'branch')

        return "<{} on {}>".format(self.__class__.__name__, ' and '.join(filters))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.process import results
from buildbot.steps import gitdiffinfo
from buildbot.test.fake.remotecommand import Expect
from buildbot.test.fake.remotecommand import ExpectShell
from buildbot.test.util import steps
from buildbot.test.util.misc import TestReactorMixin

try:
    import unidiff
except ImportError:
    unidiff = None


class TestDiffInfo(steps.BuildStepMixin, TestReactorMixin, unittest.TestCase):
    if not unidiff:
        skip = 'unidiff is required for GitDiffInfo tests'

    def setUp(self):
        self.setUpTestReactor()
        return self.setUpBuildStep()

    def tearDown(self):
        return self.tearDownBuildStep()

    def test_merge_base_failure(self):
        self.setupStep(gitdiffinfo.GitDiffInfo())
        self.expectCommands(
            ExpectShell(workdir='wkdir', command=['git', 'merge-base', 'HEAD', 'master'])
            + Expect.log('stdio-merge-base', stderr='fatal: Not a valid object name')
            + 128)
        self.expect_log_file_stderr('stdio-merge-base', 'fatal: Not a valid object name')
        self.expectOutcome(result=results.FAILURE, state_string="GitDiffInfo (failure)")
        return self.runStep()

    def test_diff_failure(self):
        self.setupStep(gitdiffinfo.GitDiffInfo())
        self.expectCommands(
            ExpectShell(workdir='wkdir', command=['git', 'merge-base', 'HEAD', 'master'])
            + Expect.log('stdio-merge-base', stdout='1234123412341234')
            + 0,
            ExpectShell(workdir='wkdir',
                        command=['git', 'diff', '--no-prefix', '-U0', '1234123412341234', 'HEAD'])
            + Expect.log('stdio-diff', stderr='fatal: ambiguous argument')
            + 1,
            )
        self.expectLogfile('stdio-merge-base', '1234123412341234')
        self.expect_log_file_stderr('stdio-diff', 'fatal: ambiguous argument')
        self.expectOutcome(result=results.FAILURE, state_string="GitDiffInfo (failure)")
        return self.runStep()

    def test_empty_diff(self):
        self.setupStep(gitdiffinfo.GitDiffInfo())
        self.expectCommands(
            ExpectShell(workdir='wkdir', command=['git', 'merge-base', 'HEAD', 'master'])
            + Expect.log('stdio-merge-base', stdout='1234123412341234')
            + 0,
            ExpectShell(workdir='wkdir',
                        command=['git', 'diff', '--no-prefix', '-U0', '1234123412341234', 'HEAD'])
            + Expect.log('stdio-diff', stdout='')
            + 0,
            )
        self.expectLogfile('stdio-merge-base', '1234123412341234')
        self.expect_log_file_stderr('stdio-diff', '')
        self.expectOutcome(result=results.SUCCESS, state_string="GitDiffInfo")
        self.expect_build_data('diffinfo-master', b'[]', 'GitDiffInfo')
        return self.runStep()

    def test_complex_diff(self):
        self.setupStep(gitdiffinfo.GitDiffInfo())
        self.expectCommands(
            ExpectShell(workdir='wkdir', command=['git', 'merge-base', 'HEAD', 'master'])
            + Expect.log('stdio-merge-base', stdout='1234123412341234')
            + 0,
            ExpectShell(workdir='wkdir',
                        command=['git', 'diff', '--no-prefix', '-U0', '1234123412341234', 'HEAD'])
            + Expect.log('stdio-diff', stdout='''\
diff --git file1 file1
deleted file mode 100644
index 42f90fd..0000000
--- file1
+++ /dev/null
@@ -1,3 +0,0 @@
-line11
-line12
-line13
diff --git file2 file2
index c337bf1..1cb02b9 100644
--- file2
+++ file2
@@ -4,0 +5,3 @@ line24
+line24n
+line24n2
+line24n3
@@ -15,0 +19,3 @@ line215
+line215n
+line215n2
+line215n3
diff --git file3 file3
new file mode 100644
index 0000000..632e269
--- /dev/null
+++ file3
@@ -0,0 +1,3 @@
+line31
+line32
+line33
''')
            + 0,
            )
        self.expectLogfile('stdio-merge-base', '1234123412341234')
        self.expectOutcome(result=results.SUCCESS, state_string="GitDiffInfo")

        diff_info = (
            b'[{"source_file": "file1", "target_file": "/dev/null", ' +
            b'"is_binary": false, "is_rename": false, ' +
            b'"hunks": [{"ss": 1, "sl": 3, "ts": 0, "tl": 0}]}, ' +
            b'{"source_file": "file2", "target_file": "file2", ' +
            b'"is_binary": false, "is_rename": false, ' +
            b'"hunks": [{"ss": 4, "sl": 0, "ts": 5, "tl": 3}, ' +
            b'{"ss": 15, "sl": 0, "ts": 19, "tl": 3}]}, ' +
            b'{"source_file": "/dev/null", "target_file": "file3", ' +
            b'"is_binary": false, "is_rename": false, ' +
            b'"hunks": [{"ss": 0, "sl": 0, "ts": 1, "tl": 3}]}]')
        self.expect_build_data('diffinfo-master', diff_info, 'GitDiffInfo')
        return self.runStep()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import jinja2

from twisted.internet import defer

from buildbot import util
from buildbot.process.properties import Properties
from buildbot.process.results import CANCELLED
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.process.results import statusToString
from buildbot.reporters import utils


def get_detected_status_text(mode, results, previous_results):
    if results == FAILURE:
        if ('change' in mode or 'problem' in mode) and previous_results is not None \
                and previous_results != FAILURE:
            text = "new failure"
        else:
            text = "failed build"
    elif results == WARNINGS:
        text = "problem in the build"
    elif results == SUCCESS:
        if "change" in mode and previous_results is not None and previous_results != results:
            text = "restored build"
        else:
            text = "passing build"
    elif results == EXCEPTION:
        text = "build exception"
    else:
        text = "{} build".format(statusToString(results))

    return text


def get_message_summary_text(build, results):
    t = build['state_string']
    if t:
        t = ": " + t
    else:
        t = ""

    if results == SUCCESS:
        text = "Build succeeded!"
    elif results == WARNINGS:
        text = "Build Had Warnings{}".format(t)
    elif results == CANCELLED:
        text = "Build was cancelled"
    else:
        text = "BUILD FAILED{}".format(t)

    return text


def get_message_source_stamp_text(source_stamps):
    text = ""

    for ss in source_stamps:
        source = ""

        if ss['branch']:
            source += "[branch {}] ".format(ss['branch'])

        if ss['revision']:
            source += str(ss['revision'])
        else:
            source += "HEAD"

        if ss['patch'] is not None:
            source += " (plus patch)"

        discriminator = ""
        if ss['codebase']:
            discriminator = " '{}'".format(ss['codebase'])

        text += "Build Source Stamp{}: {}\n".format(discriminator, source)

    return text


def get_projects_text(source_stamps, master):
    projects = set()

    for ss in source_stamps:
        if ss['project']:
            projects.add(ss['project'])

    if not projects:
        projects = [master.config.title]

    return ', '.join(list(projects))


def create_context_for_build(mode, build, master, blamelist):
    buildset = build['buildset']
    ss_list = buildset['sourcestamps']
    results = build['results']

    if 'prev_build' in build and build['prev_build'] is not None:
        previous_results = build['prev_build']['results']
    else:
        previous_results = None

    return {
        'results': build['results'],
        'mode': mode,
        'buildername': build['builder']['name'],
        'workername': build['properties'].get('workername', ["<unknown>"])[0],
        'buildset': buildset,
        'build': build,
        'projects': get_projects_text(ss_list, master),
        'previous_results': previous_results,
        'status_detected': get_detected_status_text(mode, results, previous_results),
        'build_url': utils.getURLForBuild(master, build['builder']['builderid'], build['number']),
        'buildbot_url': master.config.buildbotURL,
        'blamelist': blamelist,
        'summary': get_message_summary_text(build, results),
        'sourcestamps': get_message_source_stamp_text(ss_list)
    }


def create_context_for_worker(master, worker):
    return {
        'buildbot_title': master.config.title,
        'buildbot_url': master.config.buildbotURL,
        'worker': worker,
    }


class MessageFormatterBase(util.ComparableMixin):

    template_type = 'plain'

    def __init__(self, ctx=None, wantProperties=True, wantSteps=False, wantLogs=False):
        if ctx is None:
            ctx = {}
        self.context = ctx
        self.wantProperties = wantProperties
        self.wantSteps = wantSteps
        self.wantLogs = wantLogs

    def buildAdditionalContext(self, master, ctx):
        pass

    @defer.inlineCallbacks
    def render_message_dict(self, master, context):
        """Generate a buildbot reporter message and return a dictionary
           containing the message body, type and subject."""

        ''' This is an informal description of what message dictionaries are expected to be
            produced. It is an internal API and expected to change even within bugfix releases, if
            needed.

            The message dictionary contains the 'body', 'type' and 'subject' keys:

              - 'subject' is a string that defines a subject of the message. It's not necessarily
                used on all reporters. It may be None.

              - 'type' must be 'plain', 'html' or 'json'.

              - 'body' is the content of the message. It may be None. The type of the data depends
                on the value of the 'type' parameter:

                - 'plain': Must be a string

                - 'html': Must be a string

                - 'json': Must be a non-encoded jsonnable value. The root element must be either
                  of dictionary, list or string. This must not change during all invocations of
                  a particular instance of the formatter.

            In case of a report being created for multiple builds (e.g. in the case of a buildset),
            the values returned by message formatter are concatenated. If this is not possible
            (e.g. if the body is a dictionary), any subsequent messages are ignored.
        '''
        yield self.buildAdditionalContext(master, context)
        context.update(self.context)

        return {
            'body': (yield self.render_message_body(context)),
            'type': self.template_type,
            'subject': (yield self.render_message_subject(context))
        }

    def render_message_body(self, context):
        return None

    def render_message_subject(self, context):
        return None

    def format_message_for_build(self, master, build, **kwargs):
        # Known kwargs keys: mode, users
        raise NotImplementedError


class MessageFormatterEmpty(MessageFormatterBase):
    def format_message_for_build(self, master, build, **kwargs):
        return {
            'body': None,
            'type': 'plain',
            'subject': None
        }


class MessageFormatterFunction(MessageFormatterBase):

    def __init__(self, function, template_type, **kwargs):
        super().__init__(**kwargs)
        self.template_type = template_type
        self._function = function

    @defer.inlineCallbacks
    def format_message_for_build(self, master, build, **kwargs):
        msgdict = yield self.render_message_dict(master, {'build': build})
        return msgdict

    def render_message_body(self, context):
        return self._function(context)

    def render_message_subject(self, context):
        return None


class MessageFormatterRenderable(MessageFormatterBase):

    template_type = 'plain'

    def __init__(self, template, subject=None):
        super().__init__()
        self.template = template
        self.subject = subject

    @defer.inlineCallbacks
    def format_message_for_build(self, master, build, **kwargs):
        msgdict = yield self.render_message_dict(master, {'build': build, 'master': master})
        return msgdict

    @defer.inlineCallbacks
    def render_message_body(self, context):
        props = Properties.fromDict(context['build']['properties'])
        props.master = context['master']

        body = yield props.render(self.template)
        return body

    @defer.inlineCallbacks
    def render_message_subject(self, context):
        props = Properties.fromDict(context['build']['properties'])
        props.master = context['master']

        body = yield props.render(self.subject)
        return body


default_body_template = '''\
The Buildbot has detected a {{ status_detected }} on builder {{ buildername }} while building {{ projects }}.
Full details are available at:
    {{ build_url }}

Buildbot URL: {{ buildbot_url }}

Worker for this Build: {{ workername }}

Build Reason: {{ build['properties'].get('reason', ["<unknown>"])[0] }}
Blamelist: {{ ", ".join(blamelist) }}

{{ summary }}

Sincerely,
 -The Buildbot
'''  # noqa pylint: disable=line-too-long


class MessageFormatterBaseJinja(MessageFormatterBase):
    compare_attrs = ['body_template', 'subject_template', 'template_type']
    subject_template = None
    template_type = 'plain'

    def __init__(self, template=None, subject=None, template_type=None, **kwargs):
        if template is None:
            template = default_body_template

        self.body_template = jinja2.Template(template)

        if subject is not None:
            self.subject_template = jinja2.Template(subject)

        if template_type is not None:
            self.template_type = template_type

        super().__init__(**kwargs)

    def buildAdditionalContext(self, master, ctx):
        pass

    def render_message_body(self, context):
        return self.body_template.render(context)

    def render_message_subject(self, context):
        if self.subject_template is None:
            return None
        return self.subject_template.render(context)


class MessageFormatter(MessageFormatterBaseJinja):
    @defer.inlineCallbacks
    def format_message_for_build(self, master, build, users=None, mode=None):
        ctx = create_context_for_build(mode, build, master, users)
        msgdict = yield self.render_message_dict(master, ctx)
        return msgdict


default_missing_template = '''\
The Buildbot working for '{{buildbot_title}}' has noticed that the worker named {{worker.name}} went away.

It last disconnected at {{worker.last_connection}}.

{% if 'admin' in worker['workerinfo'] %}
The admin on record (as reported by WORKER:info/admin) was {{worker.workerinfo.admin}}.
{% endif %}

Sincerely,
 -The Buildbot
'''  # noqa pylint: disable=line-too-long


class MessageFormatterMissingWorker(MessageFormatterBaseJinja):
    template_filename = 'missing_mail.txt'

    def __init__(self, template=None, **kwargs):
        if template is None:
            template = default_missing_template
        super().__init__(template=template, **kwargs)

    @defer.inlineCallbacks
    def formatMessageForMissingWorker(self, master, worker):
        ctx = create_context_for_worker(master, worker)
        msgdict = yield self.render_message_dict(master, ctx)
        return msgdict



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# This is a static resource type and set of endpoints used as common data by
# tests.
from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types

testData = {
    13: {'id': 13, 'info': 'ok', 'success': True, 'tags': []},
    14: {'id': 14, 'info': 'failed', 'success': False, 'tags': []},
    15: {'id': 15, 'info': 'warned', 'success': True, 'tags': ['a', 'b', ]},
    16: {'id': 16, 'info': 'skipped', 'success': True, 'tags': ['a']},
    17: {'id': 17, 'info': 'ignored', 'success': True, 'tags': []},
    18: {'id': 18, 'info': 'unexp', 'success': False, 'tags': []},
    19: {'id': 19, 'info': 'todo', 'success': True, 'tags': []},
    20: {'id': 20, 'info': 'error', 'success': False, 'tags': []},
}


class TestsEndpoint(base.Endpoint):
    isCollection = True
    pathPatterns = """
    /tests
    /test
    """
    rootLinkName = 'tests'

    def get(self, resultSpec, kwargs):
        # results are sorted by ID for test stability
        return defer.succeed(sorted(testData.values(), key=lambda v: v['id']))


class RawTestsEndpoint(base.Endpoint):
    isCollection = False
    isRaw = True
    pathPatterns = "/rawtest"

    def get(self, resultSpec, kwargs):
        return defer.succeed({
            "filename": "test.txt",
            "mime-type": "text/test",
            'raw': 'value'
        })


class FailEndpoint(base.Endpoint):
    isCollection = False
    pathPatterns = "/test/fail"

    def get(self, resultSpec, kwargs):
        return defer.fail(RuntimeError('oh noes'))


class TestEndpoint(base.Endpoint):
    isCollection = False
    pathPatterns = """
    /tests/n:testid
    /test/n:testid
    """

    def get(self, resultSpec, kwargs):
        if kwargs['testid'] == 0:
            return None
        return defer.succeed(testData[kwargs['testid']])

    def control(self, action, args, kwargs):
        if action == "fail":
            return defer.fail(RuntimeError("oh noes"))
        return defer.succeed({'action': action, 'args': args, 'kwargs': kwargs})


class Test(base.ResourceType):
    name = "test"
    plural = "tests"
    endpoints = [TestsEndpoint, TestEndpoint, FailEndpoint, RawTestsEndpoint]
    keyFields = ["id"]

    class EntityType(types.Entity):
        id = types.Integer()
        info = types.String()
        success = types.Boolean()
        tags = types.List(of=types.String())
    entityType = EntityType(name)


graphql_schema = """
# custom scalar types for buildbot data model
scalar Date   # stored as utc unix timestamp
scalar Binary # arbitrary data stored as base85
scalar JSON  # arbitrary json stored as string, mainly used for properties values
type Query {
  tests(id: Int,
   id__contains: Int,
   id__eq: Int,
   id__ge: Int,
   id__gt: Int,
   id__le: Int,
   id__lt: Int,
   id__ne: Int,
   info: String,
   info__contains: String,
   info__eq: String,
   info__ge: String,
   info__gt: String,
   info__le: String,
   info__lt: String,
   info__ne: String,
   success: Boolean,
   success__contains: Boolean,
   success__eq: Boolean,
   success__ge: Boolean,
   success__gt: Boolean,
   success__le: Boolean,
   success__lt: Boolean,
   success__ne: Boolean,
   tags: String,
   tags__contains: String,
   tags__eq: String,
   tags__ge: String,
   tags__gt: String,
   tags__le: String,
   tags__lt: String,
   tags__ne: String,
   order: String,
   limit: Int,
   offset: Int): [Test]!
  test(id: Int): Test
}
type Test {
  id: Int!
  info: String!
  success: Boolean!
  tags: [String]!
}
"""
